"""MIT License

Copyright (c) 2019-2021 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import asyncio
import logging
from typing import Any, Dict, TYPE_CHECKING, Tuple, Optional

import aiohttp

import wavelink
from .utils import MISSING

if TYPE_CHECKING:
    from .pool import Node

__all__ = ("Websocket",)

logger: logging.Logger = logging.getLogger(__name__)


class Websocket:
    def __init__(self, *, node: Node, session: aiohttp.ClientSession = MISSING):
        self.node: Node = node

        self.websocket: Optional[aiohttp.ClientWebSocketResponse] = None
        if session is MISSING:
            session = aiohttp.ClientSession()
        self.session: aiohttp.ClientSession = session
        self.listener: Optional[asyncio.Task] = None

        self.host: str = f'{"https://" if self.node._https else "http://"}{self.node.host}:{self.node.port}'
        self.ws_host: str = f"ws://{self.node.host}:{self.node.port}"

    @property
    def headers(self) -> Dict[str, Any]:
        return {
            "Authorization": self.node._password,
            "User-Id": str(self.node.bot.user.id),
            "Client-Name": "WaveLink",
            'Resume-Key': self.node.resume_key
        }

    def is_connected(self) -> bool:
        return self.websocket is not None and not self.websocket.closed

    async def connect(self) -> None:
        if self.is_connected():
            assert isinstance(self.websocket, aiohttp.ClientWebSocketResponse)
            await self.websocket.close(
                code=1006, message=b"WaveLink: Attempting reconnection."
            )

        host = self.host if self.node._https else self.ws_host

        try:
            self.websocket = await self.session.ws_connect(
                host, headers=self.headers, heartbeat=self.node._heartbeat
            )
        except Exception as error:
            if (
                isinstance(error, aiohttp.WSServerHandshakeError)
                and error.status == 401
            ):
                logger.error(f"\nAuthorization Failed for Node: {self.node}\n")
            else:
                logger.error(f"Connection Failure: {error}")

            return

        if self.listener is None:
            self.listener = asyncio.create_task(self.listen())

        if self.is_connected():
            self.dispatch("node_ready", self.node)
            logger.debug(f"Connection established...{self.node.__repr__()}")

            resume = {
                "op": "configureResuming",
                "key": f"{self.node.resume_key}",
                "timeout": 60
            }
            await self.send(**resume)

    async def listen(self) -> None:
        backoff = wavelink.Backoff(base=1, maximum_time=60, maximum_tries=None)

        while True:
            assert isinstance(self.websocket, aiohttp.ClientWebSocketResponse)
            msg = await self.websocket.receive()

            if msg.type is aiohttp.WSMsgType.CLOSED:
                logger.debug(f"Websocket Closed: {msg.extra}")

                retry = backoff.calculate()

                logger.warning(f"\nRetrying connection in <{retry}> seconds...\n")

                await asyncio.sleep(retry)

                if not self.is_connected():
                    await self.connect()
            else:
                logger.debug(f"Received Payload:: <{msg.data}>")

                if msg.data == 1011:
                    # Lavalink encountered an internal error which can not be fixed...
                    # Consider updating Lavalink...
                    logger.error('Internal Lavalink Error encountered. Terminating WaveLink without retries.'
                                 'Consider updating your Lavalink Server.')

                    self.listener.cancel()
                    return

                asyncio.create_task(self.process_data(msg.json()))

    async def process_data(self, data: Dict[str, Any]) -> None:
        op = data.get("op", None)
        if not op:
            return

        if op == "stats":
            self.node.stats = wavelink.Stats(self.node, data)
            return

        try:
            player = self.node.get_player(self.node.bot.get_guild(
                int(data["guildId"])))  # type: ignore
        except KeyError:
            return

        if player is None:
            return

        if op == 'event':
            event, payload = await self._get_event_payload(data['type'], data)
            logger.debug(f'op: event:: {data}')
            
            if event == 'track_end':
                player._source = None

            self.dispatch(event, player, **payload)

        elif op == "playerUpdate":
            logger.debug(f"op: playerUpdate:: {data}")
            try:
                await player.update_state(data)
            except KeyError:
                pass

    async def _get_event_payload(
        self, name: str, data: Dict[str, Any]
    ) -> Tuple[str, Dict[str, Any]]:
        event = "event"
        payload = {}

        if name == "WebSocketClosedEvent":
            event = "websocket_closed"
            payload["reason"] = data.get("reason")
            payload["code"] = data.get("code")

        if name.startswith("Track"):
            base64_ = data.get('track')
            track = await self.node.build_track(cls=wavelink.Track, identifier=base64_)

            payload["track"] = track

            if name == "TrackEndEvent":
                event = "track_end"
                payload["reason"] = data.get("reason")

            elif name == "TrackStartEvent":
                event = "track_start"

            elif name == "TrackExceptionEvent":
                event = "track_exception"
                payload["error"] = data.get("error")

            elif name == "TrackStuckEvent":
                event = "track_stuck"
                threshold = data.get("thresholdMs")
                if isinstance(threshold, str):
                    payload["threshold"] = int(threshold)

        return event, payload

    def dispatch(self, event, *args: Any, **kwargs: Any) -> None:
        self.node.bot.dispatch(f"wavelink_{event}", *args, **kwargs)

    async def send(self, **data: Any) -> None:
        if self.is_connected():
            assert isinstance(self.websocket, aiohttp.ClientWebSocketResponse)
            logger.debug(f"Sending Payload:: {data}")

            data_str = self.node._dumps(data)
            if isinstance(data_str, bytes):
                # Some JSON libraries serialize to bytes
                # Yet Lavalink does not support binary websockets
                # So we need to decode. In the future, maybe
                # self._websocket.send_bytes could be used
                # if Lavalink ever implements it
                data_str = data_str.decode("utf-8")

            await self.websocket.send_str(data_str)



from typing import Any


__all__ = ("MISSING",)


class _MissingSentinel:
    def __repr__(self) -> str:
        return "MISSING"

    def __bool__(self) -> bool:
        return False


MISSING: Any = _MissingSentinel()



"""MIT License

Copyright (c) 2019-2022 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import abc
import collections
from typing import Any, Dict, List, Optional, Tuple


__all__ = (
    "BaseFilter",
    "Equalizer",
    "Karaoke",
    "Timescale",
    "Tremolo",
    "Vibrato",
    "Rotation",
    "Distortion",
    "ChannelMix",
    "LowPass",
    "Filter",
)


class BaseFilter(abc.ABC):

    def __init__(self, name: Optional[str] = None) -> None:
        self.name: str = name or "Unknown"

    def __repr__(self) -> str:
        return f"<wavelink.BaseFilter name={self.name}>"

    @property
    @abc.abstractmethod
    def _payload(self) -> Any:
        raise NotImplementedError


class Equalizer(BaseFilter):
    """An equalizer filter.

    Parameters
    ----------
    name: str
        The name of this filter. Can be used to differentiate between equalizer filters.
    bands: List[Dict[str, int]]
        A list of equalizer bands, each item is a dictionary with "band" and "gain" keys.
    """

    def __init__(
        self,
        name: str = "CustomEqualizer",
        *,
        bands: List[Tuple[int, float]]
    ) -> None:
        super().__init__(name=name)

        if any((band, gain) for band, gain in bands if band < 0 or band > 15 or gain < -0.25 or gain > 1.0):
            raise ValueError("Equalizer bands must be between 0 and 15 and gains between -0.25 and 1.0")

        _dict = collections.defaultdict(float)
        _dict.update(bands)

        self.bands = [{"band": band, "gain": _dict[band]} for band in range(15)]

    def __repr__(self) -> str:
        return f"<wavelink.Equalizer name={self.name}>"

    @property
    def _payload(self) -> List[Dict[str, float]]:
        return self.bands

    @classmethod
    def flat(cls) -> Equalizer:
        """A flat equalizer."""

        bands = [
            (0, 0.0), (1, 0.0), (2, 0.0), (3, 0.0), (4, 0.0),
            (5, 0.0), (6, 0.0), (7, 0.0), (8, 0.0), (9, 0.0),
            (10, 0.0), (11, 0.0), (12, 0.0), (13, 0.0), (14, 0.0)
        ]
        return cls(name="Flat EQ", bands=bands)

    @classmethod
    def boost(cls) -> Equalizer:
        """A boost equalizer."""

        bands = [
            (0, -0.075), (1, 0.125), (2, 0.125), (3, 0.1), (4, 0.1),
            (5, .05), (6, 0.075), (7, 0.0), (8, 0.0), (9, 0.0),
            (10, 0.0), (11, 0.0), (12, 0.125), (13, 0.15), (14, 0.05)
        ]
        return cls(name="Boost EQ", bands=bands)

    @classmethod
    def metal(cls) -> Equalizer:
        """A metal equalizer."""

        bands = [
            (0, 0.0), (1, 0.1), (2, 0.1), (3, 0.15), (4, 0.13),
            (5, 0.1), (6, 0.0), (7, 0.125), (8, 0.175), (9, 0.175),
            (10, 0.125), (11, 0.125), (12, 0.1), (13, 0.075), (14, 0.0)
        ]

        return cls(name="Metal EQ", bands=bands)

    @classmethod
    def piano(cls) -> Equalizer:
        """A piano equalizer."""

        bands = [
            (0, -0.25), (1, -0.25), (2, -0.125), (3, 0.0),
            (4, 0.25), (5, 0.25), (6, 0.0), (7, -0.25), (8, -0.25),
            (9, 0.0), (10, 0.0), (11, 0.5), (12, 0.25), (13, -0.025)
        ]
        return cls(name="Piano EQ", bands=bands)


class Karaoke(BaseFilter):
    """
    A Karaoke filter.

    The default values provided for all the parameters will play the track normally.

    Parameters
    ----------
    level: float
        How much of an effect this filter should have.
    mono_level: float
        How much of an effect this filter should have on mono tracks.
    filter_band: float
        The band this filter should target.
    filter_width: float
        The width of the band this filter should target.
    """

    def __init__(
        self,
        *,
        level: float = 1.0,
        mono_level: float = 1.0,
        filter_band: float = 220.0,
        filter_width: float = 100.0
    ) -> None:
        super().__init__(name="Karaoke")

        self.level: float = level
        self.mono_level: float = mono_level
        self.filter_band: float = filter_band
        self.filter_width: float = filter_width

    def __repr__(self) -> str:
        return f"<wavelink.Karaoke level={self.level}, mono_level={self.mono_level}, " \
               f"filter_band={self.filter_band}, filter_width={self.filter_width}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "level":       self.level,
            "monoLevel":   self.mono_level,
            "filterBand":  self.filter_band,
            "filterWidth": self.filter_width
        }


class Timescale(BaseFilter):
    """A timescale filter.

    Increases or decreases the speed, pitch, and/or rate of tracks.

    The default values provided for ``speed``, ``pitch`` and ``rate`` will play the track normally.

    Parameters
    ----------
    speed: float
        A multiplier for the track playback speed. Should be more than or equal to 0.0.
    pitch: float
        A multiplier for the track pitch. Should be more than or equal to 0.0.
    rate: float
        A multiplier for the track rate (pitch + speed). Should be more than or equal to 0.0.
    """

    def __init__(
        self,
        *,
        speed: float = 1.0,
        pitch: float = 1.0,
        rate: float = 1.0,
    ) -> None:

        if speed < 0:
            raise ValueError("'speed' must be more than or equal to 0.")
        if pitch < 0:
            raise ValueError("'pitch' must be more than or equal to 0.")
        if rate < 0:
            raise ValueError("'rate' must be more than or equal to 0.")

        super().__init__(name="Timescale")

        self.speed: float = speed
        self.pitch: float = pitch
        self.rate: float = rate

    def __repr__(self) -> str:
        return f"<wavelink.Timescale speed={self.speed}, pitch={self.pitch}, rate={self.rate}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "speed": self.speed,
            "pitch": self.pitch,
            "rate":  self.rate,
        }


class Tremolo(BaseFilter):
    """A tremolo filter.

    Creates a shuddering effect by quickly changing the volume.

    The default values provided for ``frequency`` and ``depth`` will play the track normally.

    Parameters
    ----------
    frequency: float
        How quickly the volume should change. Should be more than 0.0.
    depth: float
        How much the volume should change. Should be more than 0.0 and less than or equal to 1.0.
    """

    def __init__(
        self,
        *,
        frequency: float = 2.0,
        depth: float = 0.5
    ) -> None:

        if frequency < 0:
            raise ValueError("'frequency' must be more than 0.0.")
        if not 0 < depth <= 1:
            raise ValueError("'depth' must be more than 0.0 and less than or equal to 1.0.")

        super().__init__(name="Tremolo")

        self.frequency: float = frequency
        self.depth: float = depth

    def __repr__(self) -> str:
        return f"<wavelink.Tremolo frequency={self.frequency}, depth={self.depth}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "frequency": self.frequency,
            "depth":     self.depth
        }


class Vibrato(BaseFilter):
    """A vibrato filter.

    Creates a vibrating effect by quickly changing the pitch.

    The default values provided for ``frequency`` and ``depth`` will play the track normally.

    Parameters
    ----------
    frequency: float
        How quickly the pitch should change. Should be more than 0.0 and less than or equal to 14.0.
    depth: float
        How much the pitch should change. Should be more than 0.0 and less than or equal to 1.0.
    """

    def __init__(
        self,
        *,
        frequency: float = 2.0,
        depth: float = 0.5
    ) -> None:

        if not 0 < frequency <= 14:
            raise ValueError("'frequency' must be more than 0.0 and less than or equal to 14.0.")
        if not 0 < depth <= 1:
            raise ValueError("'depth' must be more than 0.0 and less than or equal to 1.0.")

        super().__init__(name="Vibrato")

        self.frequency: float = frequency
        self.depth: float = depth

    def __repr__(self) -> str:
        return f"<wavelink.Vibrato frequency={self.frequency}, depth={self.depth}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "frequency": self.frequency,
            "depth":     self.depth
        }


class Rotation(BaseFilter):
    """A rotation filter.

    Rotates the audio around stereo channels which can be used to create a 3D effect.

    The default value provided for ``speed`` will play the track normally.

    Parameters
    ----------
    speed: float
        The speed at which the audio should rotate.
    """

    def __init__(self, speed: float = 5) -> None:
        super().__init__(name="Rotation")

        self.speed: float = speed

    def __repr__(self) -> str:
        return f"<wavelink.Rotation speed={self.speed}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "rotationHz": self.speed,
        }


class Distortion(BaseFilter):
    """A distortion filter."""

    def __init__(
        self,
        *,
        sin_offset: float = 0.0,
        sin_scale: float = 1.0,
        cos_offset: float = 0.0,
        cos_scale: float = 1.0,
        tan_offset: float = 0.0,
        tan_scale: float = 1.0,
        offset: float = 0.0,
        scale: float = 1.0
    ) -> None:
        super().__init__(name="Distortion")

        self.sin_offset: float = sin_offset
        self.sin_scale: float = sin_scale
        self.cos_offset: float = cos_offset
        self.cos_scale: float = cos_scale
        self.tan_offset: float = tan_offset
        self.tan_scale: float = tan_scale
        self.offset: float = offset
        self.scale: float = scale

    def __repr__(self) -> str:
        return f"<wavelink.Distortion " \
               f"sin_offset={self.sin_offset}, " \
               f"sin_scale={self.sin_scale}, cos_offset={self.cos_offset}, " \
               f"cos_scale={self.cos_scale}, tan_offset={self.tan_offset}, " \
               f"tan_scale={self.tan_scale}, offset={self.offset}, " \
               f"scale={self.scale}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "sinOffset": self.sin_offset,
            "sinScale":  self.sin_scale,
            "cosOffset": self.cos_offset,
            "cosScale":  self.cos_scale,
            "tanOffset": self.tan_offset,
            "tanScale":  self.tan_scale,
            "offset":    self.offset,
            "scale":     self.scale
        }


class ChannelMix(BaseFilter):
    """A channel mix filter.

    Allows you to control what channel audio from the track is actually played on.

    Setting `left_to_left` and `right_to_right` to 1.0 will result in no change.
    Setting all channels to 0.5 will result in all channels receiving the same audio.

    The default values provided for all the parameters will play the track normally.

    Parameters
    ----------
    left_to_left: float
        The "percentage" of audio from the left channel that should actually get played on the left channel.
    left_to_right: float
        The "percentage" of audio from the left channel that should play on the right channel.
    right_to_left: float
        The "percentage" of audio from the right channel that should actually get played on the right channel.
    right_to_right: float
        The "percentage" of audio from the right channel that should play on the left channel.
    """

    def __init__(
        self,
        *,
        left_to_left: float = 1.0,
        left_to_right: float = 0.0,
        right_to_left: float = 0.0,
        right_to_right: float = 1.0,
    ) -> None:

        _all = (left_to_left, left_to_right, right_to_left, right_to_right)

        if any(value for value in _all if value < 0 or value > 1):
            raise ValueError(
                "'left_to_left', 'left_to_right', "
                "'right_to_left', and 'right_to_right' "
                "must all be between 0.0 and 1.0"
            )

        super().__init__(name="Channel Mix")

        self.left_to_left: float = left_to_left
        self.right_to_right: float = right_to_right
        self.left_to_right: float = left_to_right
        self.right_to_left: float = right_to_left

    def __repr__(self) -> str:
        return f"<wavelink.ChannelMix " \
               f"left_to_left={self.left_to_left}, " \
               f"right_to_right{self.right_to_right}, " \
               f"left_to_right={self.left_to_right}, " \
               f"right_to_left={self.right_to_left}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "leftToLeft":   self.left_to_left,
            "leftToRight":  self.left_to_right,
            "rightToLeft":  self.right_to_left,
            "rightToRight": self.right_to_right,
        }

    @classmethod
    def mono(cls) -> ChannelMix:
        """Returns a ChannelMix filter that will play the track in mono."""
        return cls(left_to_left=0.5, left_to_right=0.5, right_to_left=0.5, right_to_right=0.5)

    @classmethod
    def only_left(cls) -> ChannelMix:
        """Returns a ChannelMix filter that will only play the tracks left channel."""
        return cls(left_to_left=1, left_to_right=0, right_to_left=0, right_to_right=0)

    @classmethod
    def full_left(cls) -> ChannelMix:
        """
        Returns a ChannelMix filter that will play the tracks left and right channels together only on the left channel.
        """
        return cls(left_to_left=1, left_to_right=0, right_to_left=1, right_to_right=0)

    @classmethod
    def only_right(cls) -> ChannelMix:
        """Returns a ChannelMix filter that will only play the tracks right channel."""
        return cls(left_to_left=0, left_to_right=0, right_to_left=0, right_to_right=1)

    @classmethod
    def full_right(cls) -> ChannelMix:
        """
        Returns a ChannelMix filter that will play the tracks left and right channels together only on the right channel.
        """
        return cls(left_to_left=0, left_to_right=1, right_to_left=0, right_to_right=1)

    @classmethod
    def switch(cls) -> ChannelMix:
        """Returns a ChannelMix filter that will switch the tracks left and right channels."""
        return cls(left_to_left=0, left_to_right=1, right_to_left=1, right_to_right=0)


class LowPass(BaseFilter):
    """A low pass filter.

    Suppresses higher frequencies while allowing lower frequencies to pass through.

    The default value provided for ``smoothing`` will play the track normally.

    Parameters
    ----------
    smoothing: float
        The factor by which the filter will block higher frequencies.
    """

    def __init__(
        self,
        *,
        smoothing: float = 20
    ) -> None:
        super().__init__(name="Low Pass")

        self.smoothing: float = smoothing

    def __repr__(self) -> str:
        return f"<wavelink.LowPass smoothing={self.smoothing}>"

    @property
    def _payload(self) -> Dict[str, float]:
        return {
            "smoothing": self.smoothing,
        }


class Filter:
    """A filter that can be applied to a track.

    This filter accepts an instance of itself as a parameter which allows
    you to keep previous filters while also applying new ones or overwriting old ones.

    Parameters
    ----------
    filter: wavelink.Filter
        An instance of this filter class.
    volume: float
        A volume factor to apply to the track. 1.0 is the default, max is 5.0, min is 0.0.
        A value of 1.0 means the track will play at 100% volume.
    equalizer: wavelink.Equalizer
        An equalizer to apply to the track.
    karaoke: wavelink.Karaoke
        A karaoke filter to apply to the track.
    timescale: wavelink.Timescale
        A timescale filter to apply to the track.
    tremolo: wavelink.Tremolo
        A tremolo filter to apply to the track.
    vibrato: wavelink.Vibrato
        A vibrato filter to apply to the track.
    rotation: wavelink.Rotation
        A rotation filter to apply to the track.
    distortion: wavelink.Distortion
        A distortion filter to apply to the track.
    channel_mix: wavelink.ChannelMix
        A channel mix filter to apply to the track.
    low_pass: wavelink.LowPass
        A low pass filter to apply to the track.

    """

    def __init__(
        self,
        _filter: Optional[Filter] = None,
        /, *,
        volume: Optional[float] = None,
        equalizer: Optional[Equalizer] = None,
        karaoke: Optional[Karaoke] = None,
        timescale: Optional[Timescale] = None,
        tremolo: Optional[Tremolo] = None,
        vibrato: Optional[Vibrato] = None,
        rotation: Optional[Rotation] = None,
        distortion: Optional[Distortion] = None,
        channel_mix: Optional[ChannelMix] = None,
        low_pass: Optional[LowPass] = None
    ) -> None:

        self.filter: Optional[Filter] = _filter

        self.volume: Optional[float] = volume
        self.equalizer: Optional[Equalizer] = equalizer
        self.karaoke: Optional[Karaoke] = karaoke
        self.timescale: Optional[Timescale] = timescale
        self.tremolo: Optional[Tremolo] = tremolo
        self.vibrato: Optional[Vibrato] = vibrato
        self.rotation: Optional[Rotation] = rotation
        self.distortion: Optional[Distortion] = distortion
        self.channel_mix: Optional[ChannelMix] = channel_mix
        self.low_pass: Optional[LowPass] = low_pass

    def __repr__(self) -> str:
        return f"<wavelink.Filter volume={self.volume}, equalizer={self.equalizer}, " \
               f"karaoke={self.karaoke}, timescale={self.timescale}, tremolo={self.tremolo}, " \
               f"vibrato={self.vibrato}, rotation={self.rotation}, distortion={self.distortion}, " \
               f"channel_mix={self.channel_mix}, low_pass={self.low_pass}>"

    @property
    def _payload(self) -> Dict[str, Any]:

        payload = self.filter._payload.copy() if self.filter else {}

        if self.volume:
            payload["volume"] = self.volume
        if self.equalizer:
            payload["equalizer"] = self.equalizer._payload
        if self.karaoke:
            payload["karaoke"] = self.karaoke._payload
        if self.timescale:
            payload["timescale"] = self.timescale._payload
        if self.tremolo:
            payload["tremolo"] = self.tremolo._payload
        if self.vibrato:
            payload["vibrato"] = self.vibrato._payload
        if self.rotation:
            payload["rotation"] = self.rotation._payload
        if self.distortion:
            payload["distortion"] = self.distortion._payload
        if self.channel_mix:
            payload["channelMix"] = self.channel_mix._payload
        if self.low_pass:
            payload["lowPass"] = self.low_pass._payload

        return payload



"""MIT License

Copyright (c) 2019-2021 PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
from collections import deque
from copy import copy
from typing import (
    AsyncIterator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from . import abc
from .errors import *
from .types.queue import Queue as QueueBase

__all__ = (
    "Queue",
    "WaitQueue",
)


QT = TypeVar("QT", bound=QueueBase)


class Queue(Iterable[abc.Playable], Generic[QT]):
    """Basic Queue implementation for Playable objects.

    .. warning::
        This Queue class only accepts Playable objects. E.g YouTubeTrack, SoundCloudTrack.

    Parameters
    ----------
    max_size: Optional[int]
        The maximum allowed tracks in the Queue. If None, no maximum is used. Defaults to None.
    """

    __slots__ = ("max_size", "_queue", "_overflow")

    def __init__(
        self,
        max_size: Optional[int] = None,
        *,
        overflow: bool = True,
        queue_cls: Type[QT] = deque,
    ):
        self.max_size: Optional[int] = max_size
        self._queue: QT = queue_cls()  # type: ignore
        self._overflow: bool = overflow

    def __str__(self) -> str:
        """String showing all Playable objects appearing as a list."""
        return str(list(f"'{t}'" for t in self))

    def __repr__(self) -> str:
        """Official representation with max_size and member count."""
        return (
            f"<{self.__class__.__name__} max_size={self.max_size} members={self.count}>"
        )

    def __bool__(self) -> bool:
        """Treats the queue as a bool, with it evaluating True when it contains members."""
        return bool(self.count)

    def __call__(self, item: abc.Playable) -> None:
        """Allows the queue instance to be called directly in order to add a member."""
        self.put(item)

    def __len__(self) -> int:
        """Return the number of members in the queue."""
        return self.count

    def __getitem__(self, index: int) -> abc.Playable:
        """Returns a member at the given position.

        Does not remove item from queue.
        """
        if not isinstance(index, int):
            raise ValueError("'int' type required.'")

        return self._queue[index]

    def __setitem__(self, index: int, item: abc.Playable):
        """Inserts an item at given position."""
        if not isinstance(index, int):
            raise ValueError("'int' type required.'")

        self.put_at_index(index, item)

    def __delitem__(self, index: int) -> None:
        """Delete item at given position."""
        self._queue.__delitem__(index)

    def __iter__(self) -> Iterator[abc.Playable]:
        """Iterate over members in the queue.

        Does not remove items when iterating.
        """
        return self._queue.__iter__()

    def __reversed__(self) -> Iterator[abc.Playable]:
        """Iterate over members in reverse order."""
        return self._queue.__reversed__()

    def __contains__(self, item: abc.Playable) -> bool:
        """Check if an item is a member of the queue."""
        return item in self._queue

    def __add__(self, other: Iterable[abc.Playable]) -> Queue[QT]:
        """Return a new queue containing all members.

        The new queue will have the same max_size as the original.
        """
        if not isinstance(other, Iterable):
            raise TypeError(f"Adding with the '{type(other)}' type is not supported.")

        new_queue = self.copy()
        new_queue.extend(other)
        return new_queue

    def __iadd__(self, other: Union[Iterable[abc.Playable], abc.Playable]) -> Queue:
        """Add items to queue."""
        if isinstance(other, abc.Playable):
            self.put(other)
            return self

        if isinstance(other, Iterable):
            self.extend(other)
            return self

        raise TypeError(f"Adding '{type(other)}' type to the queue is not supported.")

    def _get(self) -> abc.Playable:
        return self._queue.popleft()

    def _drop(self) -> abc.Playable:
        return self._queue.pop()

    def _index(self, item: abc.Playable) -> int:
        return self._queue.index(item)

    def _put(self, item: abc.Playable) -> None:
        self._queue.append(item)

    def _insert(self, index: int, item: abc.Playable) -> None:
        self._queue.insert(index, item)

    @staticmethod
    def _check_playable(item: abc.Playable) -> abc.Playable:
        if not isinstance(item, abc.Playable):
            raise TypeError("Only Playable objects are supported.")

        return item

    @classmethod
    def _check_playable_container(cls, iterable: Iterable) -> List[abc.Playable]:
        iterable = list(iterable)
        for item in iterable:
            cls._check_playable(item)

        return iterable

    @property
    def count(self) -> int:
        """Returns queue member count."""
        return len(self._queue)

    @property
    def is_empty(self) -> bool:
        """Returns True if queue has no members."""
        return not bool(self.count)

    @property
    def is_full(self) -> bool:
        """Returns True if queue item count has reached max_size."""
        return False if self.max_size is None else self.count >= self.max_size

    def get(self) -> abc.Playable:
        """Return next immediately available item in queue if any.

        Raises QueueEmpty if no items in queue.
        """
        if self.is_empty:
            raise QueueEmpty("No items in the queue.")

        return self._get()

    def pop(self) -> abc.Playable:
        """Return item from the right end side of the queue.

        Raises QueueEmpty if no items in queue.
        """
        if self.is_empty:
            raise QueueEmpty("No items in the queue.")

        return self._queue.pop()

    def find_position(self, item: abc.Playable) -> int:
        """Find the position a given item within the queue.

        Raises ValueError if item is not in queue.
        """
        return self._index(self._check_playable(item))

    def put(self, item: abc.Playable) -> None:
        """Put the given item into the back of the queue."""
        if self.is_full:
            if not self._overflow:
                raise QueueFull(f"Queue max_size of {self.max_size} has been reached.")

            self._drop()

        return self._put(self._check_playable(item))

    def put_at_index(self, index: int, item: abc.Playable) -> None:
        """Put the given item into the queue at the specified index."""
        if self.is_full:
            if not self._overflow:
                raise QueueFull(f"Queue max_size of {self.max_size} has been reached.")

            self._drop()

        return self._insert(index, self._check_playable(item))

    def put_at_front(self, item: abc.Playable) -> None:
        """Put the given item into the front of the queue."""
        return self.put_at_index(0, item)

    def extend(self, iterable: Iterable[abc.Playable], *, atomic: bool = True) -> None:
        """
        Add the members of the given iterable to the end of the queue.

        If atomic is set to True, no tracks will be added upon any exceptions.

        If atomic is set to False, as many tracks will be added as possible.

        When overflow is enabled for the queue, `atomic=True` won't prevent dropped items.
        """
        if atomic:
            iterable = self._check_playable_container(iterable)

            if not self._overflow and self.max_size is not None:
                new_len = len(iterable)

                if (new_len + self.count) > self.max_size:
                    raise QueueFull(
                        f"Queue has {self.count}/{self.max_size} items, "
                        f"cannot add {new_len} more."
                    )

        for item in iterable:
            self.put(item)

    def copy(self) -> Queue:
        """Create a copy of the current queue including it's members."""
        new_queue = self.__class__(max_size=self.max_size)
        new_queue._queue = copy(self._queue)

        return new_queue

    def clear(self) -> None:
        """Remove all items from the queue."""
        self._queue.clear()


class WaitQueue(Queue[QT], Generic[QT]):
    """Queue for Playable objects designed for Players that allow waiting for new items with `get_wait`.

    .. note::
        WaitQueue is the default Player queue.

    Attributes
    ----------
    history: :class:`~Queue`
        A history Queue of previously played tracks.
    """

    __slots__ = ("history", "_waiters", "_finished")

    def __init__(
        self,
        max_size: Optional[int] = None,
        history_max_size: Optional[int] = None,
        history_cls=Queue[QT],
    ):
        super().__init__(max_size, overflow=False)  # type: ignore
        self.history = history_cls(history_max_size)

        self._waiters = deque()
        self._finished = asyncio.Event()
        self._finished.set()

    async def __aiter__(self) -> AsyncIterator[abc.Playable]:
        """Pops members as it iterates the queue, waiting for new members when exhausted.

        Removes items when iterating.
        """
        while True:
            yield await self.get_wait()

    def _get(self) -> abc.Playable:
        item = super()._get()
        self.history.put(item)

        return item

    def _put(self, item: abc.Playable) -> None:
        super()._put(item)
        self._wakeup_next()

    def _insert(self, index: int, item: abc.Playable) -> None:
        super()._queue.insert(index, item)
        self._wakeup_next()

    def _wakeup_next(self) -> None:
        while self._waiters:
            waiter = self._waiters.popleft()
            if not waiter.done():
                waiter.set_result(None)
                break

    async def get_wait(self) -> abc.Playable:
        """|coro|

        Return the next item in queue once available.


        .. note::
            This will wait until an item is available to be retrieved.
        """
        while self.is_empty:
            loop = asyncio.get_event_loop()
            waiter = loop.create_future()

            self._waiters.append(waiter)

            try:
                await waiter
            except:  # noqa
                waiter.cancel()

                try:
                    self._waiters.remove(waiter)
                except ValueError:  # pragma: no branch
                    pass

                if not self.is_empty and not waiter.cancelled():  # pragma: no cover
                    # something went wrong with this waiter, move on to next
                    self._wakeup_next()
                raise

        return self.get()

    async def put_wait(self, item: abc.Playable) -> None:
        """|coro|

        Put an item into the queue using await.
        """
        self._put(item)
        await asyncio.sleep(0)

    def reset(self) -> None:
        """Clears the state of all queues."""
        self.clear()
        self.history.clear()

        for waiter in self._waiters:
            waiter.cancel()

        self._waiters.clear()



from typing import Any, Iterable, Protocol


class Queue(Protocol):
    def __iter__(__self) -> Any:
        ...

    def __reversed__(__self) -> Any:
        ...

    def __getitem__(__self, __index: int) -> Any:
        ...

    def __setitem__(__self, __index: int, __value: Any) -> None:
        ...

    def __delitem__(__self, __index: int) -> None:
        ...

    def __len__(__self) -> int:
        ...

    def __iadd__(__self, __other: Iterable) -> Any:
        ...

    def append(__self, __value: Any) -> None:
        ...

    def reverse(__self) -> None:
        ...

    def extend(__self, __values: Iterable) -> None:
        ...

    def index(__self, __value: Any) -> int:
        ...

    def pop(__self) -> Any:
        ...

    def popleft(__self) -> Any:
        ...

    def remove(__self, __index: int) -> None:
        ...

    def clear(__self) -> None:
        ...

    def insert(__self, __index: int, __value: Any) -> None:
        ...
